@@ -1,4 +1,7 @@
 /* vim:set noet ts=8 sw=8 : */
+
+#define GREENLET_MODULE
+
 #include "greenlet.h"
 #include "structmember.h"
 
@@ -80,7 +83,12 @@ The running greenlet's stack_start is undefined but not NULL.
 
 /* Python 2.3 support */
 #ifndef Py_VISIT
-#define Py_VISIT(o) if (o) {if ((err = visit((PyObject *)(o), arg))) return err;}
+#define Py_VISIT(o) \
+	if (o) { \
+		if ((err = visit((PyObject *)(o), arg))) { \
+			return err; \
+		} \
+	}
 #endif /* !Py_VISIT */
 
 #ifndef Py_CLEAR
@@ -94,14 +102,27 @@ The running greenlet's stack_start is undefined but not NULL.
 	} while (0)
 #endif /* !Py_CLEAR */
 
+/* Python <= 2.5 support */
+#if PY_MAJOR_VERSION < 3
+#ifndef Py_REFCNT
+#  define Py_REFCNT(ob) (((PyObject *) (ob))->ob_refcnt)
+#endif
+#ifndef Py_TYPE
+#  define Py_TYPE(ob)   (((PyObject *) (ob))->ob_type)
+#endif
+#endif
+
+extern PyTypeObject PyGreenlet_Type;
+
 /* The current greenlet in this thread state (holds a reference) */
-static PyGreenlet* ts_current;
+static PyGreenlet* ts_current = NULL;
 /* Holds a reference to the switching-from stack during the slp switch */
-static PyGreenlet* ts_origin;
+static PyGreenlet* ts_origin = NULL;
 /* Holds a reference to the switching-to stack during the slp switch */
-static PyGreenlet* ts_target;
+static PyGreenlet* ts_target = NULL;
 /* NULL if error, otherwise args tuple to pass around during slp switch */
-static PyObject* ts_passaround;
+static PyObject* ts_passaround_args = NULL;
+static PyObject* ts_passaround_kwargs = NULL;
 
 /***********************************************************/
 /* Thread-aware routines, switching global variables when needed */
@@ -143,7 +164,7 @@ static PyGreenlet* green_create_main(void)
 	}
 
 	/* create the main greenlet for this thread */
-	gmain = (PyGreenlet*) PyType_GenericAlloc(&PyGreen_Type, 0);
+	gmain = (PyGreenlet*) PyType_GenericAlloc(&PyGreenlet_Type, 0);
 	if (gmain == NULL)
 		return NULL;
 	gmain->stack_start = (char*) 1;
@@ -196,7 +217,7 @@ static int green_updatecurrent(void)
 
 static PyObject* green_statedict(PyGreenlet* g)
 {
-	while (!PyGreen_STARTED(g))
+	while (!PyGreenlet_STARTED(g))
 		g = g->parent;
 	return g->run_info;
 }
@@ -282,7 +303,7 @@ static int slp_save_state(char* stackref)
 #define SLP_SAVE_STATE(stackref, stsizediff)		\
   stackref += STACK_MAGIC;				\
   if (slp_save_state((char*)stackref)) return -1;	\
-  if (!PyGreen_ACTIVE(ts_target)) return 1;		\
+  if (!PyGreenlet_ACTIVE(ts_target)) return 1;		\
   stsizediff = ts_target->stack_start - (char*)stackref
 
 #define SLP_RESTORE_STATE()			\
@@ -300,9 +321,9 @@ static int slp_save_state(char* stackref)
 
 /* This is a trick to prevent the compiler from inlining or
    removing the frames */
-int (*_PyGreen_slp_switch) (void);
-int (*_PyGreen_switchstack) (void);
-void (*_PyGreen_initialstub) (void*);
+int (*_PyGreenlet_slp_switch) (void);
+int (*_PyGreenlet_switchstack) (void);
+void (*_PyGreenlet_initialstub) (void*);
 
 static int g_switchstack(void)
 {
@@ -310,8 +331,9 @@ static int g_switchstack(void)
 	   that must be set before:
 	   - ts_current: current greenlet (holds a reference)
 	   - ts_target: greenlet to switch to
-	   - ts_passaround: NULL if PyErr_Occurred(),
+	   - ts_passaround_args: NULL if PyErr_Occurred(),
 	             else a tuple of args sent to ts_target (holds a reference)
+	   - ts_passaround_kwargs: same as ts_passaround_args
 	*/
 	int err;
 	{   /* save state */
@@ -320,10 +342,13 @@ static int g_switchstack(void)
 		ts_current->top_frame = tstate->frame;
 	}
 	ts_origin = ts_current;
-	err = _PyGreen_slp_switch();
+	err = _PyGreenlet_slp_switch();
 	if (err < 0) {   /* error */
-		Py_XDECREF(ts_passaround);
-		ts_passaround = NULL;
+		Py_XDECREF(ts_passaround_args);
+		ts_passaround_args = NULL;
+
+		Py_XDECREF(ts_passaround_kwargs);
+		ts_passaround_kwargs = NULL;
 	}
 	else {
 		PyThreadState* tstate = PyThreadState_GET();
@@ -337,50 +362,80 @@ static int g_switchstack(void)
 	return err;
 }
 
-static PyObject* g_switch(PyGreenlet* target, PyObject* args)
+static PyObject *
+g_switch(PyGreenlet* target, PyObject* args, PyObject* kwargs)
 {
-	/* _consumes_ a reference to the args tuple,
+	/* _consumes_ a reference to the args tuple and kwargs dict,
 	   and return a new tuple reference */
 
 	/* check ts_current */
 	if (!STATE_OK) {
 		Py_DECREF(args);
+		Py_XDECREF(kwargs);
 		return NULL;
 	}
 	if (green_statedict(target) != ts_current->run_info) {
 		PyErr_SetString(PyExc_GreenletError,
 				"cannot switch to a different thread");
 		Py_DECREF(args);
+		Py_XDECREF(kwargs);
 		return NULL;
 	}
-	ts_passaround = args;
+
+	ts_passaround_args = args;
+	ts_passaround_kwargs = kwargs;
 
 	/* find the real target by ignoring dead greenlets,
 	   and if necessary starting a greenlet. */
 	while (1) {
-		if (PyGreen_ACTIVE(target)) {
+		if (PyGreenlet_ACTIVE(target)) {
 			ts_target = target;
-			_PyGreen_switchstack();
-			return ts_passaround;
+			_PyGreenlet_switchstack();
+			break;
 		}
-		if (!PyGreen_STARTED(target)) {
+		if (!PyGreenlet_STARTED(target)) {
 			void* dummymarker;
 			ts_target = target;
-			_PyGreen_initialstub(&dummymarker);
-			return ts_passaround;
+			_PyGreenlet_initialstub(&dummymarker);
+			break;
 		}
 		target = target->parent;
 	}
+
+	/* We need to figure out what values to pass to the target greenlet
+	   based on the arguments that have been passed to greenlet.switch(). If
+	   switch() was just passed an arg tuple, then we'll just return that.
+	   If only keyword arguments were passed, then we'll pass the keyword
+	   argument dict. Otherwise, we'll create a tuple of (args, kwargs) and
+	   return both. */
+	if (ts_passaround_kwargs == NULL
+	    || PyDict_Size(ts_passaround_kwargs) == 0)
+	{
+		return ts_passaround_args;
+	}
+	else if (PySequence_Length(ts_passaround_args) == 0)
+	{
+		return ts_passaround_kwargs;
+	}
+	else
+	{
+		PyObject *tuple = PyTuple_New(2);
+		PyTuple_SetItem(tuple, 0, ts_passaround_args);
+		PyTuple_SetItem(tuple, 1, ts_passaround_kwargs);
+		return tuple;
+	}
 }
 
-static PyObject *g_handle_exit(PyObject *result)
+static PyObject *
+g_handle_exit(PyObject *result)
 {
-	if (result == NULL &&
-	    PyErr_ExceptionMatches(PyExc_GreenletExit)) {
+	if (result == NULL && PyErr_ExceptionMatches(PyExc_GreenletExit))
+	{
 		/* catch and ignore GreenletExit */
 		PyObject *exc, *val, *tb;
 		PyErr_Fetch(&exc, &val, &tb);
-		if (val == NULL) {
+		if (val == NULL)
+		{
 			Py_INCREF(Py_None);
 			val = Py_None;
 		}
@@ -388,14 +443,19 @@ static PyObject *g_handle_exit(PyObject *result)
 		Py_DECREF(exc);
 		Py_XDECREF(tb);
 	}
-	if (result != NULL) {
+	if (result != NULL)
+	{
 		/* package the result into a 1-tuple */
-		PyObject* r = result;
+		PyObject *r = result;
 		result = PyTuple_New(1);
 		if (result)
+		{
 			PyTuple_SET_ITEM(result, 0, r);
+		}
 		else
+		{
 			Py_DECREF(r);
+		}
 	}
 	return result;
 }
@@ -408,8 +468,11 @@ static void g_initialstub(void* mark)
 	/* ts_target.run is the object to call in the new greenlet */
 	PyObject* run = PyObject_GetAttrString((PyObject*) ts_target, "run");
 	if (run == NULL) {
-		Py_XDECREF(ts_passaround);
-		ts_passaround = NULL;
+		Py_XDECREF(ts_passaround_args);
+		ts_passaround_args = NULL;
+
+		Py_XDECREF(ts_passaround_kwargs);
+		ts_passaround_kwargs = NULL;
 		return;
 	}
 	/* now use run_info to store the statedict */
@@ -421,13 +484,16 @@ static void g_initialstub(void* mark)
 	/* start the greenlet */
 	ts_target->stack_start = NULL;
 	ts_target->stack_stop = (char*) mark;
-	if (ts_current->stack_start == NULL)    /* ts_current is dying */
+	if (ts_current->stack_start == NULL) {
+		/* ts_current is dying */
 		ts_target->stack_prev = ts_current->stack_prev;
-	else
+	}
+	else {
 		ts_target->stack_prev = ts_current;
+	}
 	ts_target->top_frame = NULL;
 	ts_target->recursion_depth = PyThreadState_GET()->recursion_depth;
-	err = _PyGreen_switchstack();
+	err = _PyGreenlet_switchstack();
 	/* returns twice!
 	   The 1st time with err=1: we are in the new greenlet
 	   The 2nd time with err=0: back in the caller's greenlet
@@ -435,25 +501,30 @@ static void g_initialstub(void* mark)
 	if (err == 1) {
 		/* in the new greenlet */
 		PyObject* args;
+		PyObject* kwargs;
 		PyObject* result;
 		PyGreenlet* ts_self = ts_current;
 		ts_self->stack_start = (char*) 1;  /* running */
 
-		args = ts_passaround;
+		args = ts_passaround_args;
+		kwargs = ts_passaround_kwargs;
 		if (args == NULL)    /* pending exception */
 			result = NULL;
 		else {
-			/* call g.run(*args) */
-			result = PyEval_CallObject(run, args);
+			/* call g.run(*args, **kwargs) */
+			result = PyEval_CallObjectWithKeywords(
+				run, args, kwargs);
 			Py_DECREF(args);
+			Py_XDECREF(kwargs);
 		}
 		Py_DECREF(run);
 		result = g_handle_exit(result);
+
 		/* jump back to parent */
 		ts_self->stack_start = NULL;  /* dead */
-		g_switch(ts_self->parent, result);
+		g_switch(ts_self->parent, result, NULL);
 		/* must not return from here! */
-		PyErr_WriteUnraisable((PyObject*) ts_self);
+		PyErr_WriteUnraisable((PyObject *) ts_self);
 		Py_FatalError("greenlets cannot continue");
 	}
 	/* back in the parent */
@@ -483,12 +554,12 @@ static PyObject* green_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
 static int green_setrun(PyGreenlet* self, PyObject* nparent, void* c);
 static int green_setparent(PyGreenlet* self, PyObject* nparent, void* c);
 
-static int green_init(PyGreenlet *self, PyObject *args, PyObject *kwds)
+static int green_init(PyGreenlet *self, PyObject *args, PyObject *kwargs)
 {
 	PyObject *run = NULL;
 	PyObject* nparent = NULL;
 	static char *kwlist[] = {"run", "parent", 0};
-	if (!PyArg_ParseTupleAndKeywords(args, kwds, "|OO:green", kwlist,
+	if (!PyArg_ParseTupleAndKeywords(args, kwargs, "|OO:green", kwlist,
 					 &run, &nparent))
 		return -1;
 
@@ -513,13 +584,14 @@ static int kill_greenlet(PyGreenlet* self)
 		   because the 'parent' field chain would hold a
 		   reference */
 		PyObject* result;
-		if (!STATE_OK)
+		if (!STATE_OK) {
 			return -1;
+		}
 		Py_INCREF(ts_current);
 		self->parent = ts_current;
 		/* Send the greenlet a GreenletExit exception. */
 		PyErr_SetNone(PyExc_GreenletExit);
-		result = g_switch(self, NULL);
+		result = g_switch(self, NULL, NULL);
 		if (result == NULL)
 			return -1;
 		Py_DECREF(result);
@@ -568,7 +640,7 @@ static int green_is_gc(PyGreenlet* self)
 static int green_clear(PyGreenlet* self)
 {
 	printf("green_clear %p run_info %p\n", self, self->run_info);
-	if (PyGreen_ACTIVE(self))
+	if (PyGreenlet_ACTIVE(self))
 		return kill_greenlet(self);
 	return 0;
 }
@@ -583,14 +655,12 @@ static void green_dealloc(PyGreenlet* self)
 	PyObject_GC_UnTrack((PyObject *)self);
 	Py_TRASHCAN_SAFE_BEGIN(self);
 #endif /* GREENLET_USE_GC */
-
-	Py_XDECREF(self->parent);
-	self->parent = NULL;
-	if (PyGreen_ACTIVE(self)) {
+	Py_CLEAR(self->parent);
+	if (PyGreenlet_ACTIVE(self)) {
 		/* Hacks hacks hacks copied from instance_dealloc() */
 		/* Temporarily resurrect the greenlet. */
-		assert(self->ob_refcnt == 0);
-		self->ob_refcnt = 1;
+		assert(Py_REFCNT(self) == 0);
+		Py_REFCNT(self) = 1;
 		/* Save the current exception, if any. */
 		PyErr_Fetch(&error_type, &error_value, &error_traceback);
 		if (kill_greenlet(self) < 0) {
@@ -602,9 +672,9 @@ static void green_dealloc(PyGreenlet* self)
 		/* Undo the temporary resurrection; can't use DECREF here,
 		 * it would cause a recursive call.
 		 */
-		assert(self->ob_refcnt > 0);
-		--self->ob_refcnt;
-		if (self->ob_refcnt == 0 && PyGreen_ACTIVE(self)) {
+		assert(Py_REFCNT(self) > 0);
+		--Py_REFCNT(self);
+		if (Py_REFCNT(self) == 0 && PyGreenlet_ACTIVE(self)) {
 			/* Not resurrected, but still not dead!
 			   XXX what else should we do? we complain. */
 			PyObject* f = PySys_GetObject("stderr");
@@ -616,25 +686,25 @@ static void green_dealloc(PyGreenlet* self)
 			}
 			Py_INCREF(self);   /* leak! */
 		}
-		if (self->ob_refcnt != 0) {
+		if (Py_REFCNT(self) != 0) {
 			/* Resurrected! */
-			int refcnt = self->ob_refcnt;
+			int refcnt = Py_REFCNT(self);
 			_Py_NewReference((PyObject*) self);
 #ifdef GREENLET_USE_GC
 			PyObject_GC_Track((PyObject *)self);
 #endif
-			self->ob_refcnt = refcnt;
+			Py_REFCNT(self) = refcnt;
 #ifdef COUNT_ALLOCS
-			--self->ob_type->tp_frees;
-			--self->ob_type->tp_allocs;
+			--Py_TYPE(self)->tp_frees;
+			--Py_TYPE(self)->tp_allocs;
 #endif /* COUNT_ALLOCS */
 			goto green_dealloc_end;
 		}
 	}
 	if (self->weakreflist != NULL)
 		PyObject_ClearWeakRefs((PyObject *) self);
-	Py_XDECREF(self->run_info);
-	self->ob_type->tp_free((PyObject*) self);
+	Py_CLEAR(self->run_info);
+	Py_TYPE(self)->tp_free((PyObject*) self);
 green_dealloc_end:
 #ifdef GREENLET_USE_GC
 	Py_TRASHCAN_SAFE_END(self);
@@ -656,72 +726,97 @@ static PyObject* single_result(PyObject* results)
 }
 
 static PyObject *
-throw_greenlet(PyGreenlet* self, PyObject* typ, PyObject* val, PyObject* tb)
+throw_greenlet(PyGreenlet *self, PyObject *typ, PyObject *val, PyObject *tb)
 {
 	/* Note: _consumes_ a reference to typ, val, tb */
 	PyObject *result = NULL;
 	PyErr_Restore(typ, val, tb);
-	if (PyGreen_STARTED(self) && !PyGreen_ACTIVE(self)) {
+	if (PyGreenlet_STARTED(self) && !PyGreenlet_ACTIVE(self))
+	{
 		/* dead greenlet: turn GreenletExit into a regular return */
 		result = g_handle_exit(result);
 	}
-	return single_result(g_switch(self, result));
+	return single_result(g_switch(self, result, NULL));
 }
+
 PyDoc_STRVAR(green_switch_doc,
-"switch(*args)\n\
+"switch(*args, **kwargs)\n\
 \n\
 Switch execution to this greenlet.\n\
 \n\
 If this greenlet has never been run, then this greenlet\n\
-will be switched to using the body of self.run(*args).\n\
+will be switched to using the body of self.run(*args, **kwargs).\n\
 \n\
 If the greenlet is active (has been run, but was switch()'ed\n\
 out before leaving its run function), then this greenlet will\n\
 be resumed and the return value to its switch call will be\n\
 None if no arguments are given, the given argument if one\n\
-argument is given, or the args tuple if multiple arguments\n\
-are given\n\
+argument is given, or the args tuple and keyword args dict if\n\
+multiple arguments are given.\n\
 \n\
 If the greenlet is dead, or is the current greenlet then this\n\
-function will simply return the args using the same rules as\n\
+function will simply return the arguments using the same rules as\n\
 above.");
 
-static PyObject* green_switch(PyGreenlet* self, PyObject* args)
+static PyObject* green_switch(
+	PyGreenlet* self,
+	PyObject* args,
+	PyObject* kwargs)
 {
 	Py_INCREF(args);
-	return single_result(g_switch(self, args));
+	Py_XINCREF(kwargs);
+	return single_result(g_switch(self, args, kwargs));
 }
 
-#ifndef PyExceptionClass_Check      /* Python < 2.5 */
-# define PyExceptionClass_Check     PyClass_Check
+/* Macros required to support Python < 2.6 for green_throw() */
+#ifndef PyExceptionClass_Check
+#  define PyExceptionClass_Check    PyClass_Check
 #endif
-#ifndef PyExceptionInstance_Check   /* Python < 2.5 */
-# define PyExceptionInstance_Check  PyInstance_Check
+#ifndef PyExceptionInstance_Check
+#  define PyExceptionInstance_Check PyInstance_Check
 #endif
-#ifndef PyExceptionInstance_Class   /* Python < 2.5 */
-# define PyExceptionInstance_Class(x)				\
-		((PyObject*)((PyInstanceObject*)(x))->in_class)
+#ifndef PyExceptionInstance_Class
+#  define PyExceptionInstance_Class(x) \
+	((PyObject *) ((PyInstanceObject *)(x))->in_class)
 #endif
 
-PyDoc_STRVAR(throw_doc,
-"throw(typ[,val[,tb]]) -> raise exception in greenlet, return value passed "
-"when switching back");
+PyDoc_STRVAR(green_throw_doc,
+"Switches execution to the greenlet ``g``, but immediately raises the\n"
+"given exception in ``g``.  If no argument is provided, the exception\n"
+"defaults to ``greenlet.GreenletExit``.  The normal exception\n"
+"propagation rules apply, as described above.  Note that calling this\n"
+"method is almost equivalent to the following::\n"
+"\n"
+"    def raiser():\n"
+"        raise typ, val, tb\n"
+"    g_raiser = greenlet(raiser, parent=g)\n"
+"    g_raiser.switch()\n"
+"\n"
+"except that this trick does not work for the\n"
+"``greenlet.GreenletExit`` exception, which would not propagate\n"
+"from ``g_raiser`` to ``g``.\n");
 
-static PyObject* green_throw(PyGreenlet* self, PyObject* args)
+static PyObject *
+green_throw(PyGreenlet *self, PyObject *args)
 {
 	PyObject *typ = PyExc_GreenletExit;
 	PyObject *val = NULL;
 	PyObject *tb = NULL;
-	
-	if (!PyArg_ParseTuple(args, "|OOO:throw", &typ, &val, &tb))
-		return NULL;
 
-	/* First, check the traceback argument, replacing None with
-	   NULL. */
+	if (!PyArg_ParseTuple(args, "|OOO:throw", &typ, &val, &tb))
+	{
+		return NULL;
+	}
+
+	/* First, check the traceback argument, replacing None, with NULL */
 	if (tb == Py_None)
+	{
 		tb = NULL;
-	else if (tb != NULL && !PyTraceBack_Check(tb)) {
-		PyErr_SetString(PyExc_TypeError,
+	}
+	else if (tb != NULL && !PyTraceBack_Check(tb))
+	{
+		PyErr_SetString(
+			PyExc_TypeError,
 			"throw() third argument must be a traceback object");
 		return NULL;
 	}
@@ -730,18 +825,22 @@ static PyObject* green_throw(PyGreenlet* self, PyObject* args)
 	Py_XINCREF(val);
 	Py_XINCREF(tb);
 
-	if (PyExceptionClass_Check(typ)) {
+	if (PyExceptionClass_Check(typ))
+	{
 		PyErr_NormalizeException(&typ, &val, &tb);
 	}
-
-	else if (PyExceptionInstance_Check(typ)) {
-		/* Raising an instance.  The value should be a dummy. */
-		if (val && val != Py_None) {
-			PyErr_SetString(PyExc_TypeError,
-			  "instance exception may not have a separate value");
+	else if (PyExceptionInstance_Check(typ))
+	{
+		/* Raising an instance. The value should be a dummy. */
+		if (val && val != Py_None)
+		{
+			PyErr_SetString(
+				PyExc_TypeError,
+				"instance exception may not have a separate value");
 			goto failed_throw;
 		}
-		else {
+		else
+		{
 			/* Normalize to raise <class>, <instance> */
 			Py_XDECREF(val);
 			val = typ;
@@ -749,17 +848,19 @@ static PyObject* green_throw(PyGreenlet* self, PyObject* args)
 			Py_INCREF(typ);
 		}
 	}
-	else {
-		/* Not something you can raise.  throw() fails. */
-		PyErr_Format(PyExc_TypeError,
-			     "exceptions must be classes, or instances, not %s",
-			     typ->ob_type->tp_name);
+	else
+	{
+		/* Not something you can raise. throw() fails. */
+		PyErr_Format(
+			PyExc_TypeError,
+			"exceptions must be classes, or instances, not %s",
+			Py_TYPE(typ)->tp_name);
 		goto failed_throw;
 	}
 
 	return throw_greenlet(self, typ, val, tb);
 
-failed_throw:
+ failed_throw:
 	/* Didn't use our arguments, so restore their original refcounts */
 	Py_DECREF(typ);
 	Py_XDECREF(val);
@@ -767,15 +868,15 @@ failed_throw:
 	return NULL;
 }
 
-static int green_nonzero(PyGreenlet* self)
+static int green_bool(PyGreenlet* self)
 {
-	return PyGreen_ACTIVE(self);
+	return PyGreenlet_ACTIVE(self);
 }
 
 static PyObject* green_getdead(PyGreenlet* self, void* c)
 {
 	PyObject* res;
-	if (PyGreen_ACTIVE(self) || !PyGreen_STARTED(self))
+	if (PyGreenlet_ACTIVE(self) || !PyGreenlet_STARTED(self))
 		res = Py_False;
 	else
 		res = Py_True;
@@ -785,7 +886,7 @@ static PyObject* green_getdead(PyGreenlet* self, void* c)
 
 static PyObject* green_getrun(PyGreenlet* self, void* c)
 {
-	if (PyGreen_STARTED(self) || self->run_info == NULL) {
+	if (PyGreenlet_STARTED(self) || self->run_info == NULL) {
 		PyErr_SetString(PyExc_AttributeError, "run");
 		return NULL;
 	}
@@ -796,7 +897,7 @@ static PyObject* green_getrun(PyGreenlet* self, void* c)
 static int green_setrun(PyGreenlet* self, PyObject* nrun, void* c)
 {
 	PyObject* o;
-	if (PyGreen_STARTED(self)) {
+	if (PyGreenlet_STARTED(self)) {
 		PyErr_SetString(PyExc_AttributeError,
 				"run cannot be set "
 				"after the start of the greenlet");
@@ -823,7 +924,7 @@ static int green_setparent(PyGreenlet* self, PyObject* nparent, void* c)
 		PyErr_SetString(PyExc_AttributeError, "can't delete attribute");
 		return -1;
 	}
-	if (!PyGreen_Check(nparent)) {
+	if (!PyGreenlet_Check(nparent)) {
 		PyErr_SetString(PyExc_TypeError, "parent must be a greenlet");
 		return -1;
 	}
@@ -836,7 +937,7 @@ static int green_setparent(PyGreenlet* self, PyObject* nparent, void* c)
 	p = self->parent;
 	self->parent = (PyGreenlet*) nparent;
 	Py_INCREF(nparent);
-	Py_DECREF(p);
+	Py_XDECREF(p);
 	return 0;
 }
 
@@ -848,63 +949,102 @@ static PyObject* green_getframe(PyGreenlet* self, void* c)
 }
 
 
-/***********************************************************/
-/* C interface */
+/*****************************************************************************
+ * C interface
+ *
+ * These are exported using the CObject API
+ */
 
-PyObject* PyGreen_New(PyObject* run, PyObject* parent)
+static PyGreenlet *
+PyGreenlet_GetCurrent(void)
 {
-	PyGreenlet* o;
-	if (!PyGreen_Check(parent)) {
-		PyErr_SetString(PyExc_TypeError, "parent must be a greenlet");
+	if (!STATE_OK) {
 		return NULL;
 	}
-	o = (PyGreenlet*) PyType_GenericAlloc(&PyGreen_Type, 0);
-	if (o == NULL)
-		return NULL;
-	Py_INCREF(run);
-	o->run_info = run;
-	Py_INCREF(parent);
-	o->parent = (PyGreenlet*) parent;
-	return (PyObject*) o;
+	Py_INCREF(ts_current);
+	return ts_current;
 }
 
-PyObject* PyGreen_Current(void)
+static int
+PyGreenlet_SetParent(PyGreenlet *g, PyGreenlet *nparent)
 {
-	if (!STATE_OK)
-		return NULL;
-	return (PyObject*) ts_current;
-}
-
-PyObject* PyGreen_Switch(PyObject* g, PyObject* value)
-{
-	PyGreenlet *self;
-	if (!PyGreen_Check(g)) {
-		PyErr_BadInternalCall();
-		return NULL;
-	}
-	self = (PyGreenlet*) g;
-	Py_XINCREF(value);
-	if (PyGreen_STARTED(self) && !PyGreen_ACTIVE(self))
-		value = g_handle_exit(value);
-	return single_result(g_switch(self, value));
-}
-
-int PyGreen_SetParent(PyObject* g, PyObject* nparent)
-{
-	if (!PyGreen_Check(g)) {
-		PyErr_BadInternalCall();
+	if (!STATE_OK) {
 		return -1;
 	}
-	return green_setparent((PyGreenlet*) g, nparent, NULL);
+
+	if (!PyGreenlet_Check(g)) {
+		PyErr_SetString(PyExc_TypeError, "parent must be a greenlet");
+		return -1;
+	}
+
+	return green_setparent((PyGreenlet*) g, (PyObject *) nparent, NULL);
 }
 
-/***********************************************************/
+static PyGreenlet *
+PyGreenlet_New(PyObject *run, PyGreenlet *parent)
+{
+	PyGreenlet* g = NULL;
 
+	g = (PyGreenlet *) PyType_GenericAlloc(&PyGreenlet_Type, 0);
+	if (g == NULL) {
+		return NULL;
+	}
+
+	if (run != NULL) {
+		Py_INCREF(run);
+		g->run_info = run;
+	}
+	if (parent == NULL) {
+		parent = PyGreenlet_GetCurrent();
+	}
+	PyGreenlet_SetParent(g, parent);
+	return g;
+}
+
+static PyObject *
+PyGreenlet_Switch(PyGreenlet *g, PyObject *args, PyObject *kwargs)
+{
+	PyGreenlet *self = (PyGreenlet *) g;
+
+	if (!PyGreenlet_Check(self)) {
+		PyErr_BadArgument();
+		return NULL;
+	}
+
+	if (args == NULL) {
+		args = Py_BuildValue("()");
+	}
+	else {
+		Py_INCREF(args);
+	}
+
+	if (kwargs != NULL && PyDict_Check(kwargs)) {
+		Py_INCREF(kwargs);
+	}
+	else {
+		kwargs = NULL;
+	}
+
+	return single_result(g_switch(self, args, kwargs));
+}
+
+static PyObject *
+PyGreenlet_Throw(PyGreenlet *self, PyObject *typ, PyObject *val, PyObject *tb)
+{
+	if (!PyGreenlet_Check(self)) {
+		PyErr_BadArgument();
+		return NULL;
+	}
+	return throw_greenlet(self, typ, val, tb);
+}
+
+/** End C API ****************************************************************/
 
 static PyMethodDef green_methods[] = {
-	{"switch", (PyCFunction)green_switch, METH_VARARGS, green_switch_doc},
-	{"throw",  (PyCFunction)green_throw,  METH_VARARGS, throw_doc},
-	{NULL,     NULL}		/* sentinel */
+    {"switch", (PyCFunction)green_switch,
+     METH_VARARGS | METH_KEYWORDS, green_switch_doc},
+    {"throw",  (PyCFunction)green_throw,  METH_VARARGS, green_throw_doc},
+    {NULL,     NULL}		/* sentinel */
 };
 
 static PyGetSetDef green_getsets[] = {
@@ -923,19 +1063,26 @@ static PyNumberMethods green_as_number = {
 	NULL,		/* nb_add */
 	NULL,		/* nb_subtract */
 	NULL,		/* nb_multiply */
+#if PY_MAJOR_VERSION < 3
 	NULL,		/* nb_divide */
+#endif
 	NULL,		/* nb_remainder */
 	NULL,		/* nb_divmod */
 	NULL,		/* nb_power */
 	NULL,		/* nb_negative */
 	NULL,		/* nb_positive */
 	NULL,		/* nb_absolute */
-	(inquiry)green_nonzero,	/* nb_nonzero */
+	(inquiry)green_bool,	/* nb_bool */
 };
 
-PyTypeObject PyGreen_Type = {
+
+PyTypeObject PyGreenlet_Type = {
+#if PY_MAJOR_VERSION >= 3
+	PyVarObject_HEAD_INIT(NULL, 0)
+#else
 	PyObject_HEAD_INIT(NULL)
 	0,					/* ob_size */
+#endif
 	"greenlet.greenlet",			/* tp_name */
 	sizeof(PyGreenlet),			/* tp_basicsize */
 	0,					/* tp_itemsize */
@@ -956,10 +1103,11 @@ PyTypeObject PyGreen_Type = {
 	0,					/* tp_setattro */
 	0,					/* tp_as_buffer*/
 	Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE | GREENLET_GC_FLAGS,	/* tp_flags */
-	"greenlet(run=None, parent=None)\n\
-Create a new greenlet object (without running it).  \"run\" is the\n\
-callable to invoke, and \"parent\" is the parent greenlet, which\n\
-defaults to the current greenlet.",		/* tp_doc */
+	"greenlet(run=None, parent=None) -> greenlet\n\n"
+	"Creates a new greenlet object (without running it).\n\n"
+	" - *run* -- The callable to invoke.\n"
+	" - *parent* -- The parent greenlet. The default is the current "
+	"greenlet.",                            /* tp_doc */
 	(traverseproc)GREENLET_tp_traverse,	/* tp_traverse */
 	(inquiry)GREENLET_tp_clear,		/* tp_clear */
 	0,					/* tp_richcompare */
@@ -982,7 +1130,6 @@ defaults to the current greenlet.",		/* tp_doc */
 };
 /* XXX need GC support */
 
-
 static PyObject* mod_getcurrent(PyObject* self)
 {
 	if (!STATE_OK)
@@ -992,94 +1139,92 @@ static PyObject* mod_getcurrent(PyObject* self)
 }
 
 static PyMethodDef GreenMethods[] = {
-	{"getcurrent", (PyCFunction)mod_getcurrent, METH_NOARGS,
-	"greenlet.getcurrent()\n\
-Returns the current greenlet (i.e. the one which called this\n\
-function)."},
-	{NULL,     NULL}	/* Sentinel */
+	{"getcurrent", (PyCFunction)mod_getcurrent, METH_NOARGS, /*XXX*/ NULL},
+	{NULL,     NULL}        /* Sentinel */
 };
 
 static char* copy_on_greentype[] = {
 	"getcurrent", "error", "GreenletExit", NULL
 };
 
-void initgreenlet(void)
-{
-	PyObject* m;
-	PyObject* greenletexit_doc;
-	PyObject* greenletexit_dict;
-	PyObject* greenleterror_doc;
-	PyObject* greenleterror_dict;
-	int error;
-	char** p;
-	_PyGreen_switchstack = g_switchstack;
-	_PyGreen_slp_switch = slp_switch;
-	_PyGreen_initialstub = g_initialstub;
-	m = Py_InitModule("greenlet", GreenMethods);
-	if (m == NULL)
-		return;
-	
-	if (PyModule_AddStringConstant(m, "__version__", "0.2") < 0)
-		return;
+#if PY_MAJOR_VERSION >= 3
+#define INITERROR return NULL
 
+static struct PyModuleDef greenlet_module_def = {
+	PyModuleDef_HEAD_INIT,
+	"greenlet",
+	NULL,
+	-1,
+	GreenMethods,
+};
+
+PyObject *
+PyInit_greenlet(void)
+#else
+#define INITERROR return
+
+void initgreenlet(void)
+#endif
+{
+	PyObject* m = NULL;
+	char** p = NULL;
+	PyObject *c_api_object;
+	static void *_PyGreenlet_API[PyGreenlet_API_pointers];
+
+	_PyGreenlet_switchstack = g_switchstack;
+	_PyGreenlet_slp_switch = slp_switch;
+	_PyGreenlet_initialstub = g_initialstub;
+
+#if PY_MAJOR_VERSION >= 3
+	m = PyModule_Create(&greenlet_module_def);
+#else
+	m = Py_InitModule("greenlet", GreenMethods);
+#endif
+	if (m == NULL)
+	{
+		INITERROR;
+	}
+
+	if (PyModule_AddStringConstant(m, "__version__", GREENLET_VERSION) < 0)
+	{
+		INITERROR;
+	}
+
+#if PY_MAJOR_VERSION >= 3
+	ts_curkey = PyUnicode_InternFromString("__greenlet_ts_curkey");
+	ts_delkey = PyUnicode_InternFromString("__greenlet_ts_delkey");
+#else
 	ts_curkey = PyString_InternFromString("__greenlet_ts_curkey");
 	ts_delkey = PyString_InternFromString("__greenlet_ts_delkey");
+#endif
 	if (ts_curkey == NULL || ts_delkey == NULL)
-		return;
-	if (PyType_Ready(&PyGreen_Type) < 0)
-		return;
-
-        greenleterror_dict = PyDict_New();
-	if (greenleterror_dict == NULL)
-		return;
-	greenleterror_doc = PyString_FromString("internal greenlet error");
-	if (greenleterror_doc == NULL) {
-                Py_DECREF(greenleterror_dict);
-		return;
-        }
-
-        error = PyDict_SetItemString(greenleterror_dict, "__doc__", greenleterror_doc);
-	Py_DECREF(greenleterror_doc);
-	if (error == -1) {
-                Py_DECREF(greenleterror_dict);
-		return;
-        }
-
-	PyExc_GreenletError = PyErr_NewException("py.magic.greenlet.error", NULL, greenleterror_dict);
-        Py_DECREF(greenleterror_dict);
+	{
+		INITERROR;
+	}
+	if (PyType_Ready(&PyGreenlet_Type) < 0)
+	{
+		INITERROR;
+	}
+	PyExc_GreenletError = PyErr_NewException("greenlet.error", NULL, NULL);
 	if (PyExc_GreenletError == NULL)
-		return;
-
-        greenletexit_dict = PyDict_New();
-	if (greenletexit_dict == NULL)
-		return;
-	greenletexit_doc = PyString_FromString("greenlet.GreenletExit\n\
-This special exception does not propagate to the parent greenlet; it\n\
-can be used to kill a single greenlet.\n");
-	if (greenletexit_doc == NULL) {
-                Py_DECREF(greenletexit_dict);
-		return;
-        }
-
-        error = PyDict_SetItemString(greenletexit_dict, "__doc__", greenletexit_doc);
-	Py_DECREF(greenletexit_doc);
-	if (error == -1) {
-                Py_DECREF(greenletexit_dict);
-		return;
-        }
-
-	PyExc_GreenletExit = PyErr_NewException("py.magic.greenlet.GreenletExit",
-						NULL, greenletexit_dict);
-        Py_DECREF(greenletexit_dict);
+	{
+		INITERROR;
+	}
+	PyExc_GreenletExit = PyErr_NewException("greenlet.GreenletExit",
+						NULL, NULL);
 	if (PyExc_GreenletExit == NULL)
-		return;
+	{
+		INITERROR;
+	}
 
 	ts_current = green_create_main();
 	if (ts_current == NULL)
-		return;
+	{
+		INITERROR;
+	}
 
-	Py_INCREF(&PyGreen_Type);
-	PyModule_AddObject(m, "greenlet", (PyObject*) &PyGreen_Type);
+	Py_INCREF(&PyGreenlet_Type);
+	PyModule_AddObject(m, "greenlet", (PyObject*) &PyGreenlet_Type);
 	Py_INCREF(PyExc_GreenletError);
 	PyModule_AddObject(m, "error", PyExc_GreenletError);
 	Py_INCREF(PyExc_GreenletExit);
@@ -1090,11 +1235,41 @@ can be used to kill a single greenlet.\n");
 	PyModule_AddObject(m, "GREENLET_USE_GC", PyBool_FromLong(0));
 #endif
 
-	/* also publish module-level data as attributes of the greentype. */
+        /* also publish module-level data as attributes of the greentype. */
 	for (p=copy_on_greentype; *p; p++) {
 		PyObject* o = PyObject_GetAttrString(m, *p);
 		if (!o) continue;
-		PyDict_SetItemString(PyGreen_Type.tp_dict, *p, o);
+		PyDict_SetItemString(PyGreenlet_Type.tp_dict, *p, o);
 		Py_DECREF(o);
 	}
+
+	/*
+	 * Expose C API
+	 */
+
+	/* types */
+	_PyGreenlet_API[PyGreenlet_Type_NUM] = (void *) &PyGreenlet_Type;
+
+	/* exceptions */
+	_PyGreenlet_API[PyExc_GreenletError_NUM] = (void *) PyExc_GreenletError;
+	_PyGreenlet_API[PyExc_GreenletExit_NUM] = (void *) PyExc_GreenletExit;
+
+	/* methods */
+	_PyGreenlet_API[PyGreenlet_New_NUM] = (void *) PyGreenlet_New;
+	_PyGreenlet_API[PyGreenlet_GetCurrent_NUM] =
+		(void *) PyGreenlet_GetCurrent;
+	_PyGreenlet_API[PyGreenlet_Throw_NUM] = (void *) PyGreenlet_Throw;
+	_PyGreenlet_API[PyGreenlet_Switch_NUM] = (void *) PyGreenlet_Switch;
+	_PyGreenlet_API[PyGreenlet_SetParent_NUM] =
+		(void *) PyGreenlet_SetParent;
+
+	c_api_object = PyCObject_FromVoidPtr((void *) _PyGreenlet_API, NULL);
+	if (c_api_object != NULL)
+	{
+		PyModule_AddObject(m, "_C_API", c_api_object);
+	}
+
+#if PY_MAJOR_VERSION >= 3
+	return m;
+#endif
 }
